"""
Implementation of GB energy prediction paper
Weike Ye, Hui Zheng, Chi Chen, Shyue Ping Ong
"A Universal Machine Learning Model for Elemental Grain Boundary Energies"
arXiv preprint arXiv: 2201.11991 (2022).
"""
